/*
 * Copyright 2015 Main Street Softworks, Inc. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modification, are
 * permitted provided that the following conditions are met:
 * 
 *  1. Redistributions of source code must retain the above copyright notice, this list of
 *     conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright notice, this list
 *     of conditions and the following disclaimer in the documentation and/or other materials
 *     provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY MAIN STREET SOFTWORKS INC ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL MAIN STREET SOFTWORKS INC OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * The views and conclusions contained in the software and documentation are those of the
 * authors and should not be interpreted as representing official policies, either expressed
 * or implied, of Main Street Softworks, Inc.
 */

package com.monetra.csv;

import java.io.ByteArrayOutputStream;

public class CsvParser
{
	String[][] mCsv;

	public CsvParser(String data) {
		parseCsv(data.getBytes(), (byte)',', (byte)'"');
	}

	private String byteArrayToString(byte[] data) {
		try {
			return new String(data, "UTF8");
		} catch (Exception e) {  }
		return null;
	}

	private byte[] byteArraySubStr(byte[] str, int start, int length) {
		if (str == null || str.length < length)
			return null;

		byte[] ret = new byte[length];
		System.arraycopy(str, start, ret, 0, length);
		return ret;
	}

	private String removeDupQuotes(byte[] str) {
		/* No quotes */
		if (byteArrayChr(str, (byte)'"') == -1)
			return byteArrayToString(str);

		ByteArrayOutputStream out_str = new ByteArrayOutputStream();
		for (int i=0; i<str.length; i++) {
			if (str[i] == (byte)'"' && i < str.length-1 && str[i+1] == (byte)'"') {
				out_str.write((byte)'"');
				i++;
			} else if (str[i] != (byte)'"') {
				out_str.write(str[i]);
			}
		}
		return byteArrayToString(out_str.toByteArray());
	}

	private int byteArrayChr(byte[] data, byte chr) {
		for (int i=0; i<data.length; i++) {
			if (data[i] == chr)
				return i;
		}
		return -1;
	}

	private byte[] byteArrayTrim(byte[] str) {
		int start = -1;
		int stop  = -1;
		int length = 0;
		int i;

		if (str == null || str.length == 0)
			return new byte[0];

		for (i=0; i<str.length; i++) {
			if (str[i] == (byte)'\r' || str[i] == (byte)'\n' ||
					str[i] == (byte)'\t' || str[i] == (byte)' ')
			{
				continue;
			}
			start = i;
			break;
		}

		if (start == -1)
			return new byte[0];

		for (i=str.length-1; i>= 0; i--) {
			if (str[i] == (byte)'\r' || str[i] == (byte)'\n' ||
					str[i] == (byte)'\t' || str[i] == (byte)' ')
			{
				continue;
			}
			stop = i;
			break;
		}

		if (stop == -1)
			return new byte[0];

		length = (stop-start)+1;
		if (length == 0)
			return new byte[0];
		if (length == str.length)
			return str;

		return byteArraySubStr(str, start, length);
	}

    private byte[][] byteArrayExplode(byte delim, byte[] data, byte quote_char, int max_sects)
    {
        Boolean on_quote;
        int beginsect;
        if (data == null || data.length == 0)
            return null;

		/* We need to first count how many lines we have */
        on_quote  = false;
        beginsect = 0;
        int num_lines = 0;
        for (int i=0; i<data.length; i++) {
            if (quote_char != 0 && data[i] == quote_char) {
				/* Doubling the quote char acts as escaping */
                if (data[i+1] == quote_char) {
                    i++;
                    continue;
                } else if (on_quote) {
                    on_quote = false;
                } else {
                    on_quote = true;
                }
            }
            if (data[i] == delim && !on_quote) {
                num_lines++;
                beginsect = i+1;
                if (max_sects != 0 && num_lines == max_sects-1)
                    break;
            }
        }
        if (beginsect < data.length)
            num_lines++;

        byte[][] ret = new byte[num_lines][];
        beginsect     = 0;
        int cnt       = 0;
        on_quote      = false;

        for (int i=0; i<data.length; i++) {
            if (quote_char != 0 && data[i] == quote_char) {
				/* Doubling the quote char acts as escaping */
                if (data[i+1] == quote_char) {
                    i++;
                    continue;
                } else if (on_quote) {
                    on_quote = false;
                } else {
                    on_quote = true;
                }
            }
            if (data[i] == delim && !on_quote) {
                ret[cnt++] = byteArraySubStr(data, beginsect, i - beginsect);
                beginsect = i + 1;
                if (max_sects != 0&& cnt == max_sects - 1)
                    break;
            }
        }
        if (beginsect < data.length) {
            ret[cnt++] = byteArraySubStr(data, beginsect, data.length - beginsect);
        }
        return ret;
    }

	private void parseCsv(byte[] data, byte delimiter, byte enclosure) {
		byte[][] lines = byteArrayExplode((byte)'\n', data, enclosure, 0);

		mCsv = new String[lines.length][];
		for (int i=0; i<lines.length; i++) {
			byte[][] cells = byteArrayExplode(delimiter, lines[i], enclosure, 0);
			mCsv[i] = new String[cells.length];
			for (int j=0; j<cells.length; j++) {
				mCsv[i][j] = removeDupQuotes(byteArrayTrim(cells[j]));
			}
		}
	}

	public static boolean isCsv(String data) {
        byte[] bdata = data.getBytes();
		for (int i=0; i<bdata.length; i++) {
			/* If hit a new line or a comma before an equal sign, must
			 * be comma delimited */
			if (bdata[i] == '\r' ||
					bdata[i] == '\n' ||
					bdata[i] == ',')
				return true;

			/* If hit an equal sign before a new line or a comma, must be
			 * key/val */
			if (bdata[i] == '=')
				return false;
		}
		/* Who knows?  Should never get here */
		return true;
	}

	public int numColumns() {
		return mCsv[0].length;
	}

	public int numRows() {
		return mCsv.length-1;
	}

	public String getHeader(int col) {
		if (col >= mCsv[0].length)
			return null;

		return mCsv[0][col];
	}

	public String getCell(int row, String col) {
		if (row+1 >= mCsv.length)
			return null;

		for (int i=0; i<mCsv[0].length; i++) {
			if (col.equalsIgnoreCase(mCsv[0][i])) {
				return mCsv[row+1][i];
			}
		}
		return null;
	}

	public String getCell(int row, int col) 	{
		if (row+1 >= mCsv.length || col >= mCsv[0].length)
			return null;

		return mCsv[row+1][col];
	}
}

