/*
 * Copyright 2015 Main Street Softworks, Inc. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modification, are
 * permitted provided that the following conditions are met:
 * 
 *  1. Redistributions of source code must retain the above copyright notice, this list of
 *     conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright notice, this list
 *     of conditions and the following disclaimer in the documentation and/or other materials
 *     provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY MAIN STREET SOFTWORKS INC ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL MAIN STREET SOFTWORKS INC OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * The views and conclusions contained in the software and documentation are those of the
 * authors and should not be interpreted as representing official policies, either expressed
 * or implied, of Main Street Softworks, Inc.
 */

package com.monetra.uniterm.sdk;

import android.app.Activity;
import android.app.Fragment;
import android.content.ComponentName;
import android.content.Context;
import android.content.Intent;
import android.content.ServiceConnection;
import android.os.AsyncTask;
import android.os.Bundle;
import android.os.IBinder;
import android.util.Pair;

import java.security.SecureRandom;
import java.util.HashMap;

import com.monetra.uniterm.service.IUnitermService;


/**
 * Integrate with Uniterm with an easy to use API that handles many common tasks.
 *
 * <p>
 * Example of how to use persistently from within an Activity:
 * </p>
 *
 * <pre>
 * {@code
 * FragmentManager fm = getFragmentManager();
 * UnitermServiceHelper mUnitermServiceHelper = (UnitermServiceHelper) fm.findFragmentByTag("ush_f1");
 * if (mUnitermServiceHelper == null) {
 *     mUnitermServiceHelper = new UnitermServiceHelper();
 *     fm.beginTransaction().add(mUnitermServiceHelper, "ush_f1").commit();
 *     fm.executePendingTransactions();
 * }
 * }
 * </pre>
 *
 * <p>
 * This allows the fragment and its data to be retained after a configuration change.
 * </p>
 *
 * <p>
 * Note: Only one txnrequest can run at a time. To run multiple transaction requests
 * use multiple helper objects.
 * </p>
 */
public class UnitermServiceHelper extends Fragment {

    private UnitermServiceHelperCallbacks mCallbacks;
    private UnitermServiceHelperImp mHelperImp;
    // Data from the currently running txnrequest so we can pull status and
    // cancel. This will also be used for synchronization so it can never be null.
    private HashMap<String, String> mRunningTxn = new HashMap<>();

    private enum RequestType {
        NONE,
        TXNREQUEST,
        TXNSTART,
        TXNFINISH,
        STATUS,
        CANCEL,
        DEVICELOAD,
        PASSTHROUGH,
        VERSION,
        BLUETOOTHLIST,
        DEVICETYPES,
        RAWKVS
    }

    /**
     * Callbacks for the available operations.
     */
    public interface UnitermServiceHelperCallbacks {
        void TxnrequestResponse(HashMap<String, String> kvs);
        void TxnstartResponse(HashMap<String, String> kvs);
        void TxnfinishResponse(HashMap<String, String> kvs);
        void TxnrequestStatusResponse(HashMap<String, String> kvs);
        void TxnrequestCancelResponse(HashMap<String, String> kvs);
        void DeviceloadResponse(HashMap<String, String> kvs);
        void PassthroughResponse(HashMap<String, String> kvs);
        void VersionResponse(HashMap<String, String> kvs);
        void BluetoothlistResponse(String csv);
        void DevicetypesResponse(String csv);
        void RawKVS(HashMap<String, String> kvs);
    }

    @Override
    public void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);

        // Retain this fragment across configuration changes.
        setRetainInstance(true);

        // Holds the service connection. Bind the service to the application context because
        // it persists for the life of the application. Binding it to an activity will cause
        // get a null reference when the activity gone. A screen rotation will destory and
        // rebuild an activity for exapmle.
        mHelperImp = new UnitermServiceHelperImp(getActivity().getApplicationContext());
    }

    @Override
    public void onAttach(Activity activity) {
        super.onAttach(activity);
        // The activity we're attached to must implement the callbacks. Link the callbacks
        // from the activity so we can use them.
        mCallbacks = (UnitermServiceHelperCallbacks) activity;
    }

    @Override
    public void onDetach() {
        super.onDetach();
        // If we're detaching from an activity we can't use it anymore because it's most likely
        // being destroyed. We'll reset the callbacks to a new activity on reattach.
        mCallbacks = null;
    }

    /** Is a txnrequest currently running.
     */
    public boolean IsTxnrequestRunning() {
        synchronized (mRunningTxn) {
            return !mRunningTxn.isEmpty();
        }
    }

    /**
     * Run a transaction request.
     *
	 * <p>
     * KVS should have at a minimum:
	 * </p>
	 * <ul>
     * <li>u_device</li>
     * <li>u_devicetype</li>
     * <li>username</li>
     * <li>password</li>
     * <li>amount</li>
     * <li>action</li>
	 * </ul>
     *
     * @param kvs        Key value pairs.
     * @param autoStatus Should status be run for the duration of the txn automatically?
     */
    public void DoTxnrequest(HashMap<String, String> kvs, boolean autoStatus) {
		// Check if a transaction is already running. This SDK is limited to
		// only allow one txntrequest to run at a time.
        synchronized (mRunningTxn) {
            if (!mRunningTxn.isEmpty()) {
                HashMap<String, String> result = new HashMap<>();
                result.put("code", "DENY");
                result.put("u_errorcode", "PENDING_TRAN");
                result.put("verbiage", "Transaction Running");
                result.put("u_id", mRunningTxn.get("u_id"));

                if (mHelperImp != null) {
                    mCallbacks.TxnrequestResponse(result);
                }
                return;
            }

            // Store information we'll need to run status or cancel.
            mRunningTxn = new HashMap<>();
            mRunningTxn.put("username", kvs.get("username"));
            mRunningTxn.put("password", kvs.get("password"));
            // u_id is necessary for status and cancel.
            // If one isn't provided use a random one.
            if (!kvs.containsKey("u_id")) {
                SecureRandom rand = new SecureRandom();
                byte bytes[] = new byte[6];
                rand.nextBytes(bytes);
                kvs.put("u_id", bytes.toString());
            }
            mRunningTxn.put("u_id", kvs.get("u_id"));
        }
        kvs.put("u_action", "txnrequest");
        mHelperImp.sendRequest(RequestType.TXNREQUEST, kvs);

        // Start running status requests in a thread until the txn is done.
        if (autoStatus) {
            statusLoop();
        }
    }

    public void DoTxnstart(HashMap<String, String> kvs, boolean autoStatus) {
        synchronized (mRunningTxn) {
            if (!mRunningTxn.isEmpty()) {
                HashMap<String, String> result = new HashMap<>();
                result.put("code", "DENY");
                result.put("u_errorcode", "PENDING_TRAN");
                result.put("verbiage", "Transaction Running");
                result.put("u_id", mRunningTxn.get("u_id"));

                if (mHelperImp != null) {
                    mCallbacks.TxnstartResponse(result);
                }
                return;
            }

            // Store information we'll need to run status or cancel.
            mRunningTxn = new HashMap<>();
            mRunningTxn.put("username", kvs.get("username"));
            mRunningTxn.put("password", kvs.get("password"));
            // u_id is necessary for status and cancel.
            // If one isn't provided use a random one.
            if (!kvs.containsKey("u_id")) {
                SecureRandom rand = new SecureRandom();
                byte bytes[] = new byte[6];
                rand.nextBytes(bytes);
                kvs.put("u_id", bytes.toString());
            }
            mRunningTxn.put("u_id", kvs.get("u_id"));
        }
        kvs.put("u_action", "txnstart");
        mHelperImp.sendRequest(RequestType.TXNSTART, kvs);
        if (autoStatus) {
            statusLoop();
        }
    }

    public void DoTxnFinish(HashMap<String, String> kvs, boolean autoStatus) {
        synchronized (mRunningTxn) {
            if (!mRunningTxn.isEmpty()) {
                kvs.put("u_id", mRunningTxn.get("u_id"));
            }
        }
        kvs.put("u_action", "txnfinish");

        mHelperImp.sendRequest(RequestType.TXNFINISH, kvs);
        if (autoStatus) {
            statusLoop();
        }
    }

    /**
     * Get the status for the currently running txnrequest.
     */
    public void DoTxnrequestStatus() {
        HashMap<String, String> kvs = new HashMap<>();

        synchronized (mRunningTxn) {
            if (mRunningTxn.isEmpty()) {
                return;
            }
            kvs.put("username", mRunningTxn.get("username"));
            kvs.put("password", mRunningTxn.get("password"));
            kvs.put("u_id", mRunningTxn.get("u_id"));
        }
        kvs.put("u_action", "status");

        mHelperImp.sendRequest(RequestType.STATUS, kvs);
    }

    /**
     * Cancel the currently running txnrequest.
     */
    public void DoTxnrequestCancel() {
        HashMap<String, String> kvs = new HashMap<>();

        synchronized (mRunningTxn) {
            if (mRunningTxn.isEmpty()) {
                return;
            }
            kvs.put("username", mRunningTxn.get("username"));
            kvs.put("password", mRunningTxn.get("password"));
            kvs.put("u_id", mRunningTxn.get("u_id"));
        }
        kvs.put("u_action", "cancel");

        mHelperImp.sendRequest(RequestType.CANCEL, kvs);
    }

    /**
     * Load the device.
     *
	 * <p>
     * KVS should have at a minimum:
	 * </p>
	 * <ul>
     * <li>u_device</li>
     * <li>u_devicetype</li>
     * <li>username</li>
     * <li>password</li>
	 * </ul>
     *
     * @param kvs   Key value pairs.
     * @param force Force loading the device.
     */
    public void DoDeviceload(HashMap<String, String> kvs, boolean force) {
        if (force && !kvs.containsKey("u_forceload")) {
            kvs.put("u_forceload", "yes");
        }
        kvs.put("u_action", "deviceload");
        mHelperImp.sendRequest(RequestType.DEVICELOAD, kvs);
    }

    /**
     * Run a passthrough transaction to Monetra.
     *
     * @param kvs Key value pairs.
     */
    public void DoPassthrough(HashMap<String, String> kvs) {
        kvs.put("u_action", "passthrough");
        mHelperImp.sendRequest(RequestType.PASSTHROUGH, kvs);
    }

    /**
     * Get the version of Uniterm.
     */
    public void DoVersion() {
        HashMap<String, String> kvs = new HashMap<>();
        kvs.put("u_action", "version");
        mHelperImp.sendRequest(RequestType.VERSION, kvs);
    }

    /**
     * Get a list of configured bluetooth devices.
     *
	 * <p>
     * This is all bluetooth devices not only payment devices.
	 * </p>
     */
    public void DoBluetoothlist() {
        HashMap<String, String> kvs = new HashMap<>();
        kvs.put("u_action", "bluetoothlist");
        mHelperImp.sendRequest(RequestType.BLUETOOTHLIST, kvs);
    }

    /**
     * Get a list of supported device types. This is limited to devices that support bluetooth.
     */
    public void DoDevicetypes() {
        HashMap<String, String> kvs = new HashMap<>();
        kvs.put("u_action", "devicetypes");
        mHelperImp.sendRequest(RequestType.DEVICETYPES, kvs);
    }

    /**
     * Send kvs to Uniterm.
     *
	 * <p>
     * This can be used to send commands that do not have a convenience function here.
     * This should only be used to fill gaps in this interface not as the sole means of
     * communicating with Uniterm. If the rest of the helper functions are not used it
     * is recommended to integrate with Uniterm directly though the aidl interface.
	 * </p>
     *
     * @param kvs
     */
    public void DoRawKVS(HashMap<String, String> kvs, boolean autoStatus) {
        mHelperImp.sendRequest(RequestType.RAWKVS, kvs);
        if (autoStatus) {
            statusLoop();
        }
    }

    private void statusLoop() {
        Thread t = new Thread() {
            public void run() {
                while (true) {
                    synchronized (mRunningTxn) {
                        if (mRunningTxn.isEmpty()) {
                            break;
                        }
                    }

                    try {
                        Thread.sleep(500);
                    } catch (Exception e) {
                        break;
                    }

                    DoTxnrequestStatus();
                }
            }
        };
        t.start();
    }

    /**
     * Handles communication with the Uniterm service.
     */
    private class UnitermServiceHelperImp {

        private IUnitermService mUnitermService = null;
        private Context mContext = null;

		/**
 		 * Initalize the service helper.
		 *
		 * @param c Context that the service will bind to. Be sure to bind to an application context
		 * or bad things could happen.
		 */
        public UnitermServiceHelperImp(Context c) {
            mContext = c;
        }

        private ServiceConnection mUnitermServiceConnection = new ServiceConnection() {
            @Override
            public void onServiceConnected(ComponentName name, IBinder service) {
                synchronized (this) {
                    mUnitermService = IUnitermService.Stub.asInterface(service);
                }
            }

            @Override
            public void onServiceDisconnected(ComponentName name) {
                synchronized (this) {
                    mUnitermService = null;
                }
            }
        };

        private synchronized void connectUniterm() {
            synchronized (this) {
                if (mUnitermService != null) {
                    return;
                }
            }

            Intent i = new Intent();
            i.setClassName("com.monetra.uniterm", "com.monetra.uniterm.service.UnitermService");
            mContext.bindService(i, mUnitermServiceConnection, Context.BIND_AUTO_CREATE);
        }

        protected void finalize() {
            if (mUnitermServiceConnection != null) {
                mContext.unbindService(mUnitermServiceConnection);
            }
            mContext = null;
        }

        /**
         * Call the callback for a given action.
         *
         * @param requestType Type of request.
         * @param out         kvs to send.
         */
        public void callCallback(RequestType requestType, HashMap<String, String> out) {
            if (mCallbacks == null)
                return;

            switch (requestType) {
                case TXNREQUEST:
                    synchronized (mRunningTxn) {
                        mRunningTxn.clear();
                    }
                    mCallbacks.TxnrequestResponse(out);
                    break;
                case TXNSTART:
                    if (!out.get("u_errorcode").equals("SUCCESS")) {
                        synchronized (mRunningTxn) {
                            mRunningTxn.clear();
                        }
                    }
                    mCallbacks.TxnstartResponse(out);
                    break;
                case TXNFINISH:
                    synchronized (mRunningTxn) {
                        mRunningTxn.clear();
                    }
                    mCallbacks.TxnfinishResponse(out);
                    break;
                case STATUS:
                    mCallbacks.TxnrequestStatusResponse(out);
                    break;
                case CANCEL:
                    mCallbacks.TxnrequestCancelResponse(out);
                    synchronized (mRunningTxn) {
                        mRunningTxn.clear();
                    }
                    break;
                case DEVICELOAD:
                    mCallbacks.DeviceloadResponse(out);
                    break;
                case PASSTHROUGH:
                    mCallbacks.PassthroughResponse(out);
                    break;
                case VERSION:
                    mCallbacks.VersionResponse(out);
                    break;
                case BLUETOOTHLIST:
                    mCallbacks.BluetoothlistResponse(out.get("DataBlock"));
                    break;
                case DEVICETYPES:
                    mCallbacks.DevicetypesResponse(out.get("DataBlock"));
                    break;
                case RAWKVS:
                    mCallbacks.RawKVS(out);
                    break;
                case NONE:
                    break;
            }
        }

        /**
         * Send a request to Uniterm asynchronously.
         */
        public void sendRequest(RequestType requestType, HashMap<String, String> params) {
            // Each request will be run in it's own thread from the AsyncTask thread pool.
            new UnitermRequest().executeOnExecutor(AsyncTask.THREAD_POOL_EXECUTOR, new Pair<>(requestType, params));
        }

        /**
         * Asynchronous task for communicating with Uniterm.
         *
		 * <p>
         * Takes a request type and kvs to send.
		 * </p>
         *
		 * <p>
         * Does not update progress because the communication with the Uniterm Service is
         * blocking. One of the reasons why the request is run though an AsyncTask.
		 * </p>
         *
		 * <p>
         * Returns to the callbacks the result kvs.
		 * </p>
         *
		 * <p>
         * Before using a callback check if the callback object is null. This could happen if
         * this taks is still running after onDestroy for both the activity and fragment has been
         * called.
		 * </p>
         *
		 * <p>
         * When callbacks are not null we don't have to worry about a race condition where
         * onDetach -> AsyncTask onPostExecute -> onAttach. The AyncTask schedules onPostExecute
         * to run on the UI thread. onDetach and onAttach will run before onPostExecute because
         * configuration change events are handled in a single message. onPostExecute will be in
         * its own message separate from the other two. onPostExecute can only run before or
         * after onDetach and onAttach not between. See http://stackoverflow.com/a/20775934
		 * </p>
         */
        private class UnitermRequest extends AsyncTask<Pair<RequestType, HashMap<String, String>>, Void, Pair<RequestType, HashMap<String, String>>> {

            /**
             * Process the request.
             *
             * @param input Pair with request type and kvs.
             * @return Pair with request type and result kvs.
             */
            @Override
            protected Pair<RequestType, HashMap<String, String>> doInBackground(Pair<RequestType, HashMap<String, String>>... input) {
                RequestType requestType = (RequestType)input[0].first;
                HashMap<String, String> params = (HashMap<String, String>)input[0].second;
                Bundle request = new Bundle();
                HashMap<String, String> out = new HashMap<String, String>();
                Bundle result;
                int count = 0;

                // Ensure we are connected to Uniterm and start an asynchronous connection
                // attempt if we're not.
                connectUniterm();

                // Add the uid from the running transactions to the output so
                // A single callback can be used for multiple helper objects
                // to differentiate what result this is for.
                if(params.containsKey("u_id")) {
                    out.put("u_id", params.get("u_id"));
                }

                // If Uniterm is not connected yet we want to wait a bit to make sure it's not
                // due to the connection attempt being asynchronous. We might not be connected
                // if we've just started up or there was an interruption in the connection.
                while (mUnitermService == null && count < 10) {
                    try {
                        Thread.sleep(500);
                    } catch (Exception e) {
                        // pass
                    }
                    count++;
                }

                synchronized (this) {
                    if (mUnitermService == null) {
                        out.put("code", "DENY");
                        out.put("verbiage", "Not connected to Uniterm Service!");
                        return new Pair<>(requestType, out);
                    }
                }

                for (String key : params.keySet()) {
                    request.putString(key, params.get(key));
                }
                try {
                    result = mUnitermService.runTrans(request);
                    for (String key : result.keySet()) {
                        out.put(key, result.getString(key));
                    }
                } catch (Exception e) {
                    out.clear();
                    out.put("code", "DENY");
                    out.put("verbiage", e.toString());
                }

                return new Pair<>(requestType, out);
            }

            @Override
            protected void onProgressUpdate(Void... ignore) {

            }

            @Override
            protected void onPostExecute(Pair<RequestType, HashMap<String, String>> out) {
                callCallback(out.first, out.second);
            }
        }
    }
}
